import numpy as np
import math

import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
import matplotlib.colors as colors
import matplotlib.cm as cmx
import matplotlib as mpl


from operator import sub
from matplotlib import gridspec
from matplotlib import rc


from matplotlib.ticker import FormatStrFormatter
from matplotlib.ticker import ScalarFormatter
from matplotlib.ticker import FuncFormatter

#Astrophysics:

#Constants:

from astropy import constants as const

import h5py

h_Planck16 = 0.6774

#Formatting


mnras_column_with_pt = np.e * 240.00 #np.e * 240.00
pt_to_inch = 0.01389
golden_mean = (np.sqrt(5.0)-1.0)/2.0 

width = pt_to_inch * mnras_column_with_pt #* 1.2
height = width * golden_mean * 1.2


#plt.style.use('dark_background')

params = {'backend' : 'qt4agg',
            'text.latex.preamble' : r"\usepackage{lmodern}",
            'text.latex.preview': True,
            'text.usetex' : True,
            'font.size' : 22,#18
            'font.family' : 'lmodern',
            'text.latex.unicode' : True,
            'figure.figsize' : (width, height),
            'figure.subplot.hspace' : 0.03,
            'axes.labelsize' : 20,#18
            'axes.labelpad': 1.,
            'xtick.labelsize' : 20,#18
            'ytick.labelsize' : 20,
            'xtick.major.pad' : 6, 
            'ytick.major.pad' : 6, #18
            'xtick.major.size' : 3.5,     # major tick size in points
          'xtick.minor.size' : 2.5,      # minor tick size in points
          'xtick.major.width'    : 1,    # major tick width in points
          'xtick.minor.width'    : 1, 
          'ytick.major.size' : 3.5,     # major tick size in points
          'ytick.minor.size' : 2.5,      # minor tick size in points
          'ytick.major.width'    : 1,    # major tick width in points
          'ytick.minor.width'    : 1, 
            'legend.fontsize' : 17,#17
            'legend.numpoints' : 1,
            'legend.scatterpoints' : 1,
            'legend.handletextpad' : 0.3,
            'legend.handlelength' : 1,
            'legend.handleheight' : 1,
            'legend.columnspacing' : 1,
            'legend.frameon' : True,
            'savefig.bbox' : 'tight',
            'savefig.format' : 'pdf',
            }

plt.rcParams.update(params) 

#Bootstrap percentiles

n_wished_percentile = 95
n_lower_percentile_bound = (100 - n_wished_percentile)/2
n_upper_percentile_bound = 50 + n_wished_percentile/2

n_bootstrap_samples = 100


def delta_sigma_signal_calculator(n_halos, n_radial_bins, min_rad, max_rad, bootresult, particle_mass, signal_core_x, signal_core_y, signal_core_z, signal_rad_x, signal_rad_y, signal_rad_z):

    n_halos_inverse = 1./n_halos

    log10_bin_size = (np.log10(max_rad) - np.log10(min_rad))/n_radial_bins

    radial_bins_edges_log10 = np.linspace(np.log10(min_rad), np.log10(max_rad), n_radial_bins + 1)

    radial_bins = 10 ** ((radial_bins_edges_log10[:-1] + radial_bins_edges_log10[1:]) / 2)
    
    signal_core = np.hstack((signal_core_x, signal_core_y, signal_core_z))

    correlation_core_joint = signal_core.sum()
    
    signal_rad = np.vstack((signal_rad_x, signal_rad_y, signal_rad_z))

    correlation_radial_bins_joint = np.sum(signal_rad, axis = 0)


    core_bootstrap = np.mean(np.array([np.sum(signal_core[bootresult], axis = 0)]), axis = 0)

    correlation_radial_bins_joint_bootstrap = np.mean(np.array([np.sum(signal_rad[bootresult, :], axis = 0)]), axis = 0)

    radial_weights = n_halos_inverse * particle_mass * 1./(math.pi * ((10. ** radial_bins_edges_log10[1:]) ** 2 - (10. ** radial_bins_edges_log10[0:n_radial_bins]) ** 2))

    sigma_signal = radial_weights * correlation_radial_bins_joint


    core_array = np.array([correlation_core_joint])

    radial_average_particle_weights_inner_outer = np.cumsum(np.concatenate((core_array, correlation_radial_bins_joint), axis=0))

    radial_average_weights = 0.5 * n_halos_inverse * particle_mass * 1./math.pi * (radial_average_particle_weights_inner_outer[0:n_radial_bins] * 1./((10. ** radial_bins_edges_log10[0:n_radial_bins]) ** 2) + radial_average_particle_weights_inner_outer[1:] * 1./(10. ** radial_bins_edges_log10[1:]) ** 2)


    delta_sigma_signal = (radial_average_weights - radial_weights * correlation_radial_bins_joint) * 0.01


     # Bootstrap


    sigma_signal_bootstrap = radial_weights * correlation_radial_bins_joint_bootstrap


    radial_average_particle_weights_inner_outer_bootstrap = np.cumsum(np.column_stack((core_bootstrap, correlation_radial_bins_joint_bootstrap)), axis = 1)

    radial_average_weights_bootstrap = 0.5 * n_halos_inverse * particle_mass * 1./math.pi * (radial_average_particle_weights_inner_outer_bootstrap[:,0:n_radial_bins] * 1./((10. ** radial_bins_edges_log10[0:n_radial_bins]) ** 2) + radial_average_particle_weights_inner_outer_bootstrap[:, 1:] * 1./(10. ** radial_bins_edges_log10[1:]) ** 2)

    delta_sigma_signal_bootstrap = (radial_average_weights_bootstrap - sigma_signal_bootstrap) * 0.01

    delta_sigma_signal_bootstrap_upper = np.percentile(delta_sigma_signal_bootstrap, n_upper_percentile_bound, interpolation = 'linear', axis = 0)

    delta_sigma_signal_bootstrap_lower = np.percentile(delta_sigma_signal_bootstrap, n_lower_percentile_bound, interpolation = 'linear', axis = 0)

    return delta_sigma_signal, delta_sigma_signal_bootstrap_upper, delta_sigma_signal_bootstrap_lower, radial_bins, delta_sigma_signal_bootstrap


def delta_sigma_signal_calculator_Illustris_db(simulation_filename, database_filename):

    fdb = h5py.File(database_filename, "r")

    galaxy_ids = fdb['/Subhalo/SubhaloIndices'][:]

    fdb.close()

    selected_galaxy_ids = np.sort(galaxy_ids, kind = 'mergesort')
    
    print("N gals tot: ", selected_galaxy_ids.shape[0])

    f = h5py.File(simulation_filename, "r")

    header_name = '/Header'

    geometry_group_name = '/ParticleCountCylinderGeometry'

    header_group = f[header_name]

    geometry_group = f[geometry_group_name]

    subhalo_id_dataset_name = '/Subgroups/SubhaloIDs'

    dm_core_dataset_name_x = '/ProjectedParticleShells/DMCoreProjectedAlongX'

    dm_radial_dataset_name_x = '/ProjectedParticleShells/DMCountsRadialShellsProjectedAlongX'

    dm_core_dataset_name_y = '/ProjectedParticleShells/DMCoreProjectedAlongY'

    dm_radial_dataset_name_y = '/ProjectedParticleShells/DMCountsRadialShellsProjectedAlongY'

    dm_core_dataset_name_z = '/ProjectedParticleShells/DMCoreProjectedAlongZ'

    dm_radial_dataset_name_z = '/ProjectedParticleShells/DMCountsRadialShellsProjectedAlongZ'


    gas_core_dataset_name_x = '/ProjectedParticleShells/GasCoreProjectedAlongX'

    gas_radial_dataset_name_x = '/ProjectedParticleShells/GasCountsRadialShellsProjectedAlongX'

    gas_core_dataset_name_y = '/ProjectedParticleShells/GasCoreProjectedAlongY'

    gas_radial_dataset_name_y = '/ProjectedParticleShells/GasCountsRadialShellsProjectedAlongY'

    gas_core_dataset_name_z = '/ProjectedParticleShells/GasCoreProjectedAlongZ'

    gas_radial_dataset_name_z = '/ProjectedParticleShells/GasCountsRadialShellsProjectedAlongZ'

    stars_and_winds_core_dataset_name_x = '/ProjectedParticleShells/StarsAndWindsCoreProjectedAlongX'

    stars_and_winds_radial_dataset_name_x = '/ProjectedParticleShells/StarsAndWindsCountsRadialShellsProjectedAlongX'

    stars_and_winds_core_dataset_name_y = '/ProjectedParticleShells/StarsAndWindsCoreProjectedAlongY'

    stars_and_winds_radial_dataset_name_y = '/ProjectedParticleShells/StarsAndWindsCountsRadialShellsProjectedAlongY'

    stars_and_winds_core_dataset_name_z = '/ProjectedParticleShells/StarsAndWindsCoreProjectedAlongZ'

    stars_and_winds_radial_dataset_name_z = '/ProjectedParticleShells/StarsAndWindsCountsRadialShellsProjectedAlongZ'

    smbh_core_dataset_name_x = '/ProjectedParticleShells/SMBHsCoreProjectedAlongX'

    smbh_radial_dataset_name_x = '/ProjectedParticleShells/SMBHsCountsRadialShellsProjectedAlongX'

    smbh_core_dataset_name_y = '/ProjectedParticleShells/SMBHsCoreProjectedAlongY'

    smbh_radial_dataset_name_y = '/ProjectedParticleShells/SMBHsCountsRadialShellsProjectedAlongY'

    smbh_core_dataset_name_z = '/ProjectedParticleShells/SMBHsCoreProjectedAlongZ'

    smbh_radial_dataset_name_z = '/ProjectedParticleShells/SMBHsCountsRadialShellsProjectedAlongZ'

    n_radial_bins = geometry_group.attrs['NRadialBins']

    rad_min = geometry_group.attrs['MinRad']

    rad_max = geometry_group.attrs['MaxRad']

    particle_mass = header_group.attrs['MPart']

    dm_core_dataset_x = f[dm_core_dataset_name_x]

    dm_radial_dataset_x = f[dm_radial_dataset_name_x]

    dm_core_dataset_y = f[dm_core_dataset_name_y]

    dm_radial_dataset_y = f[dm_radial_dataset_name_y]

    dm_core_dataset_z = f[dm_core_dataset_name_z]

    dm_radial_dataset_z = f[dm_radial_dataset_name_z]

    gas_core_dataset_x = f[gas_core_dataset_name_x]

    gas_radial_dataset_x = f[gas_radial_dataset_name_x]

    gas_core_dataset_y = f[gas_core_dataset_name_y]

    gas_radial_dataset_y = f[gas_radial_dataset_name_y]

    gas_core_dataset_z = f[gas_core_dataset_name_z]

    gas_radial_dataset_z = f[gas_radial_dataset_name_z]

    stars_and_winds_core_dataset_x = f[stars_and_winds_core_dataset_name_x]

    stars_and_winds_radial_dataset_x = f[stars_and_winds_radial_dataset_name_x]

    stars_and_winds_core_dataset_y = f[stars_and_winds_core_dataset_name_y]

    stars_and_winds_radial_dataset_y = f[stars_and_winds_radial_dataset_name_y]

    stars_and_winds_core_dataset_z = f[stars_and_winds_core_dataset_name_z]

    stars_and_winds_radial_dataset_z = f[stars_and_winds_radial_dataset_name_z]

    smbh_core_dataset_x = f[smbh_core_dataset_name_x]

    smbh_radial_dataset_x = f[smbh_radial_dataset_name_x]

    smbh_core_dataset_y = f[smbh_core_dataset_name_y]

    smbh_radial_dataset_y = f[smbh_radial_dataset_name_y]

    smbh_core_dataset_z = f[smbh_core_dataset_name_z]

    smbh_radial_dataset_z = f[smbh_radial_dataset_name_z]

    subhalo_id_dataset = f[subhalo_id_dataset_name]


    sorted_subhalo_indices = np.argsort(subhalo_id_dataset[:], kind = 'mergesort')

    mask_ids = np.in1d(subhalo_id_dataset[:][sorted_subhalo_indices], selected_galaxy_ids)

    n_selected_subhalos = subhalo_id_dataset[:][sorted_subhalo_indices[mask_ids]].shape[0]

    selected_sorted_subhalo_indices = sorted_subhalo_indices[mask_ids]


    sorted_subhalo_radial_data = np.repeat(n_radial_bins *selected_sorted_subhalo_indices, n_radial_bins) + np.tile(np.arange(n_radial_bins), n_selected_subhalos)


    dm_core_dataset_x_sorted = dm_core_dataset_x[:][selected_sorted_subhalo_indices]
    dm_core_dataset_y_sorted = dm_core_dataset_y[:][selected_sorted_subhalo_indices]
    dm_core_dataset_z_sorted = dm_core_dataset_z[:][selected_sorted_subhalo_indices]

    dm_radial_dataset_x_sorted = dm_radial_dataset_x[selected_sorted_subhalo_indices, 0:n_radial_bins]

    dm_radial_dataset_y_sorted = dm_radial_dataset_y[:][selected_sorted_subhalo_indices, 0:n_radial_bins]

    dm_radial_dataset_z_sorted = dm_radial_dataset_z[:][selected_sorted_subhalo_indices, 0:n_radial_bins]

    gas_core_dataset_x_sorted = gas_core_dataset_x[:][selected_sorted_subhalo_indices]
    gas_core_dataset_y_sorted = gas_core_dataset_y[:][selected_sorted_subhalo_indices]
    gas_core_dataset_z_sorted = gas_core_dataset_z[:][selected_sorted_subhalo_indices]

    gas_radial_dataset_x_sorted = gas_radial_dataset_x[:][selected_sorted_subhalo_indices, 0:n_radial_bins]

    gas_radial_dataset_y_sorted = gas_radial_dataset_y[:][selected_sorted_subhalo_indices, 0:n_radial_bins]

    gas_radial_dataset_z_sorted = gas_radial_dataset_z[:][selected_sorted_subhalo_indices, 0:n_radial_bins]


    stars_and_winds_core_dataset_x_sorted = stars_and_winds_core_dataset_x[:][selected_sorted_subhalo_indices]

    stars_and_winds_core_dataset_y_sorted = stars_and_winds_core_dataset_y[:][selected_sorted_subhalo_indices]

    stars_and_winds_core_dataset_z_sorted = stars_and_winds_core_dataset_z[:][selected_sorted_subhalo_indices]

    stars_and_winds_radial_dataset_x_sorted = stars_and_winds_radial_dataset_x[:][selected_sorted_subhalo_indices, 0:n_radial_bins]

    stars_and_winds_radial_dataset_y_sorted = stars_and_winds_radial_dataset_y[:][selected_sorted_subhalo_indices, 0:n_radial_bins]

    stars_and_winds_radial_dataset_z_sorted = stars_and_winds_radial_dataset_z[:][selected_sorted_subhalo_indices, 0:n_radial_bins]

    smbh_core_dataset_x_sorted = smbh_core_dataset_x[:][selected_sorted_subhalo_indices]

    smbh_core_dataset_y_sorted = smbh_core_dataset_y[:][selected_sorted_subhalo_indices]

    smbh_core_dataset_z_sorted = smbh_core_dataset_z[:][selected_sorted_subhalo_indices]

    smbh_radial_dataset_x_sorted = smbh_radial_dataset_x[:][selected_sorted_subhalo_indices, 0:n_radial_bins]

    smbh_radial_dataset_y_sorted = smbh_radial_dataset_y[:][selected_sorted_subhalo_indices, 0:n_radial_bins]

    smbh_radial_dataset_z_sorted = smbh_radial_dataset_z[:][selected_sorted_subhalo_indices, 0:n_radial_bins]


    bootresult = np.random.choice(n_selected_subhalos + n_selected_subhalos + n_selected_subhalos, size = (n_selected_subhalos + n_selected_subhalos + n_selected_subhalos, n_bootstrap_samples), replace = True)

    min_rad = rad_min / 1000
    max_rad = rad_max / 1000

    f.close()

    dm_Delta_Sigma, dm_Delta_Sigma_upper, dm_Delta_Sigma_lower, dm_Delta_Sigma_radial_bins, dm_Delta_Sigma_bootstrap = delta_sigma_signal_calculator(n_selected_subhalos + n_selected_subhalos + n_selected_subhalos, n_radial_bins, min_rad , max_rad, bootresult, particle_mass, dm_core_dataset_x_sorted, dm_core_dataset_y_sorted, dm_core_dataset_z_sorted, dm_radial_dataset_x_sorted, dm_radial_dataset_y_sorted, dm_radial_dataset_z_sorted)

    gas_Delta_Sigma, gas_Delta_Sigma_upper, gas_Delta_Sigma_lower, gas_Delta_Sigma_radial_bins, gas_Delta_Sigma_bootstrap = delta_sigma_signal_calculator(n_selected_subhalos + n_selected_subhalos + n_selected_subhalos, n_radial_bins, min_rad, max_rad, bootresult, 1., gas_core_dataset_x_sorted, gas_core_dataset_y_sorted, gas_core_dataset_z_sorted, gas_radial_dataset_x_sorted, gas_radial_dataset_y_sorted, gas_radial_dataset_z_sorted)

    stars_and_winds_Delta_Sigma, stars_and_winds_Delta_Sigma_upper, stars_and_winds_Delta_Sigma_lower, stars_and_winds_Delta_Sigma_radial_bins, stars_and_winds_Delta_Sigma_bootstrap = delta_sigma_signal_calculator(n_selected_subhalos + n_selected_subhalos + n_selected_subhalos, n_radial_bins, min_rad, max_rad, bootresult, 1., stars_and_winds_core_dataset_x_sorted, stars_and_winds_core_dataset_y_sorted, stars_and_winds_core_dataset_z_sorted, stars_and_winds_radial_dataset_x_sorted, stars_and_winds_radial_dataset_y_sorted, stars_and_winds_radial_dataset_z_sorted)

    smbh_Delta_Sigma, smbh_Delta_Sigma_upper, smbh_Delta_Sigma_lower, smbh_Delta_Sigma_radial_bins, smbh_Delta_Sigma_bootstrap = delta_sigma_signal_calculator(n_selected_subhalos + n_selected_subhalos + n_selected_subhalos, n_radial_bins, min_rad, max_rad, bootresult, 1., smbh_core_dataset_x_sorted, smbh_core_dataset_y_sorted, smbh_core_dataset_z_sorted, smbh_radial_dataset_x_sorted, smbh_radial_dataset_y_sorted, smbh_radial_dataset_z_sorted)

    return dm_Delta_Sigma + gas_Delta_Sigma + stars_and_winds_Delta_Sigma + smbh_Delta_Sigma, dm_Delta_Sigma_upper + gas_Delta_Sigma_upper + stars_and_winds_Delta_Sigma_upper + smbh_Delta_Sigma_upper, dm_Delta_Sigma_lower + gas_Delta_Sigma_lower + stars_and_winds_Delta_Sigma_lower + smbh_Delta_Sigma_lower, dm_Delta_Sigma_radial_bins, dm_Delta_Sigma_bootstrap + gas_Delta_Sigma_bootstrap + stars_and_winds_Delta_Sigma_bootstrap + smbh_Delta_Sigma_bootstrap, dm_Delta_Sigma, dm_Delta_Sigma_upper, dm_Delta_Sigma_lower, gas_Delta_Sigma, gas_Delta_Sigma_upper, gas_Delta_Sigma_lower, stars_and_winds_Delta_Sigma, stars_and_winds_Delta_Sigma_upper, stars_and_winds_Delta_Sigma_lower, smbh_Delta_Sigma, smbh_Delta_Sigma_upper, smbh_Delta_Sigma_lower



lensing_dir = "./"

lensing_filename = lensing_dir + "tng300_full_physics_lensing_profiles_per_subhalo_snapnum_99.hdf5"

database_filename = lensing_dir + "tng300_example_selection_full_physics_galaxies_with_mstar_10p11_10p11p2_Msun.hdf5"

fig0 = plt.figure(0)
gs0 = gridspec.GridSpec(1, 1, height_ratios=[1])
ax0 = plt.subplot(gs0[0])

delta_sigma_signal_tng300, delta_sigma_signal_tng300_upper, delta_sigma_signal_tng300_lower, radial_bins, _,  delta_sigma_signal_tng300_dm, delta_sigma_signal_tng300_dm_upper, delta_sigma_signal_tng300_dm_lower, delta_sigma_signal_tng300_gas, delta_sigma_signal_tng300_gas_upper, delta_sigma_signal_tng300_gas_lower, delta_sigma_signal_tng300_stars_and_winds, delta_sigma_signal_tng300_stars_and_winds_upper, delta_sigma_signal_tng300_stars_and_winds_lower, delta_sigma_signal_tng300_smbh, delta_sigma_signal_tng300_smbh_upper, delta_sigma_signal_tng300_smbh_lower =delta_sigma_signal_calculator_Illustris_db(lensing_filename, database_filename)

min_radius = radial_bins[0] * 0.99
max_radius = radial_bins[-1] * 1.01

ax0.fill_between(radial_bins, delta_sigma_signal_tng300_upper, delta_sigma_signal_tng300_lower, linewidth = 0, alpha = 0.3, color = 'xkcd:ocean blue')

ax0.plot(radial_bins, delta_sigma_signal_tng300, '-' , lw = 2, alpha = 1, color = 'xkcd:ocean blue', label=r'TNG300, tot')

print("Delta Sigma: ", delta_sigma_signal_tng300)

ax0.fill_between(radial_bins, delta_sigma_signal_tng300_dm_upper, delta_sigma_signal_tng300_dm_lower, linewidth = 0, alpha = 0.3, color = 'xkcd:mulberry')

ax0.plot(radial_bins, delta_sigma_signal_tng300_dm, '--' , lw = 2, alpha = 1, color = 'xkcd:mulberry', label=r'TNG300, DM')

ax0.fill_between(radial_bins, delta_sigma_signal_tng300_gas_upper, delta_sigma_signal_tng300_gas_lower, linewidth = 0, alpha = 0.3, color = 'xkcd:coral')

ax0.plot(radial_bins, delta_sigma_signal_tng300_gas, '-.' , lw = 2, alpha = 1, color = 'xkcd:coral', label=r'TNG300, gas')

ax0.fill_between(radial_bins, delta_sigma_signal_tng300_stars_and_winds_upper, delta_sigma_signal_tng300_stars_and_winds_lower, linewidth = 0, alpha = 0.3, color = 'xkcd:golden')

ax0.plot(radial_bins, delta_sigma_signal_tng300_stars_and_winds, ':' , lw = 2, alpha = 1, color = 'xkcd:golden', label=r'TNG300, stars')

ax0.set_yscale('log')
ax0.set_xscale('log')

ax0.set_xlim([min_radius, max_radius])

description = r'Galaxies with $%.2f < \log_{10} M_{\ast} \left [M_{\odot} \right ] < %.2f$' % (11.0, 11.2)

ax0.text(0.1, 100, description, color = 'k', bbox=dict(facecolor='none', edgecolor='xkcd:silver', boxstyle='round', alpha=0.5), size = 16)

ax0.set_xlabel(r'$r \left [ h^{-1} \, \mathrm{Mpc}\right ]$')
ax0.set_ylabel(r'$\Delta \Sigma(r) \left [ h \, M_{\odot} \, \mathrm{pc}^{-2}\right ]$')

legend = ax0.legend(loc='lower left', ncol = 1, frameon = 1, fancybox=True, framealpha=0.5)

plt.show()

#fig0.savefig("sample_fig")
